﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;

namespace NPersistence.NHibernate.Event
{
    abstract class CallbackHandler
    {
        protected Dictionary<Type, List<CallbackMethod>> callbackRegistry = new Dictionary<Type, List<CallbackMethod>>();

        private static Dictionary<Type, Type> classHierarchyRegistry = new Dictionary<Type,Type>();

        private static Dictionary<Type, CallbackHandler> internalCallbackHandlerRegistry = null;

        private bool registrationPerformed = false;

        private void registerInternalCallbackMethod(Type entityType, MethodInfo method)
        {
            //validate:
            //returns void, take no arguments, not static, event doesn't already have a callback registered
            if (!method.ReturnType.Equals(typeof(void)))
            {
                throw new PersistenceException("The callback method " + method.Name + " should be void.");
            }
            ParameterInfo[] pi = method.GetParameters();
            if (pi != null && pi.Length > 0)
            {
                throw new PersistenceException("The callback method " + method.Name + " should not take any parameters.");
            }
            if (method.IsStatic)
            {
                throw new PersistenceException("The callback method " + method.Name + " should not be static.");
            }
            List<CallbackMethod> methodList;
            if (callbackRegistry.TryGetValue(entityType, out methodList))
            {
                foreach(CallbackMethod cbMethod in methodList)
                {
                    if(cbMethod.isInternal())
                    {
                        throw new PersistenceException("The Entity " + entityType.Name + " cannot have more than than one callback method for the same event. Remove one of <" + method.Name + "> or <" + cbMethod.Name + ">");
                    }
                }
            }
            else// methodList == null
            {
                methodList = new List<CallbackMethod>();
                callbackRegistry[entityType] = methodList;
            }
            methodList.Add(new CallbackMethod(null, method));
            registrationPerformed = true;
        }

        private void registerExternalCallbackMethod(Type entityType, Object callbackObject, MethodInfo method)
        {
            //validate:
            //returns void, takes 1 argument which is from the entityType inheritence hierarchy, not static.
            if (!method.ReturnType.Equals(typeof(void)))
            {
                throw new PersistenceException("The callback method " + method.Name + " should be void.");
            }
            ParameterInfo[] pi = method.GetParameters();
            if (pi == null || pi.Length != 1)
            {
                throw new PersistenceException("The callback method " + method.Name + " should not take any parameters.");
            }
            if (!pi[0].ParameterType.IsAssignableFrom(entityType))
            {
                throw new PersistenceException("The callback method " + method.Name + " should take as parameter the entity type or a type of it's inheritence hierarchy. Instead, it takes <" + pi[0].Name + ">.");
            }
            if (method.IsStatic)
            {
                throw new PersistenceException("The callback method " + method.Name + " should not be static.");
            }
            List<CallbackMethod> methodList;
            if (!callbackRegistry.TryGetValue(entityType, out methodList))
            {
                methodList = new List<CallbackMethod>();
                callbackRegistry[entityType] = methodList;
            }
            methodList.Add(new CallbackMethod(callbackObject, method));
            registrationPerformed = true;
        }

        protected void invokeCallback(Object entity)
        {
            Type entityType = entity.GetType();
            List<CallbackMethod> methodList;
            do
            {
                if (callbackRegistry.TryGetValue(entityType, out methodList))
                {
                    foreach (CallbackMethod cbMethod in methodList)
                    {
                        cbMethod.invokeCallback(entity);
                    }
                }
                // get the parent of the entity and invoke its callback methods.
            } while (classHierarchyRegistry.TryGetValue(entityType, out entityType));
        }

        public bool hasRegistrations()
        {
            return registrationPerformed;
        }

        internal static void registerCallback(Type attributeType, Type entityType, Object callbackObject, MethodInfo method)
        {
            if (internalCallbackHandlerRegistry == null)
            {
                initCallbackHandlerRegistry();
            }
            CallbackHandler callbackHandler;
            if (internalCallbackHandlerRegistry.TryGetValue(attributeType, out callbackHandler))
            {
                if (callbackObject != null)
                {
                    callbackHandler.registerExternalCallbackMethod(entityType, callbackObject, method);
                }
                else
                {
                    callbackHandler.registerInternalCallbackMethod(entityType, method);
                }
            }
        }

        internal static void registerHierarchyCallback(Type entity, Type baseType)
        {
            classHierarchyRegistry[entity] = baseType;
        }

        private static void initCallbackHandlerRegistry()
        {
            internalCallbackHandlerRegistry = new Dictionary<Type,CallbackHandler>();
            internalCallbackHandlerRegistry[typeof(PrePersist)] = PrePersistCallbackHandler.getInstance();
            internalCallbackHandlerRegistry[typeof(PostPersist)] = PostPersistCallbackHandler.getInstance();
            internalCallbackHandlerRegistry[typeof(PostLoad)] = PostLoadCallbackHandler.getInstance();
            internalCallbackHandlerRegistry[typeof(PreUpdate)] = PreUpdateCallbackHandler.getInstance();
            internalCallbackHandlerRegistry[typeof(PostUpdate)] = PostUpdateCallbackHandler.getInstance();
            internalCallbackHandlerRegistry[typeof(PreRemove)] = PreRemoveCallbackHandler.getInstance();
            internalCallbackHandlerRegistry[typeof(PostRemove)] = PostRemoveCallbackHandler.getInstance();
        }
    }
}
