﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NPersistence.spi;
using NHibernate.Proxy;
using NHibernate.Intercept;
using NPersistence;
using System.Reflection;
using NHibernate;
using NHibernate.Collection;
using System.Runtime.Serialization;

namespace NPersistence.NHibernate.util
{
    class PersistenceUtilHelper
    {
        public static LoadState IsLoadedWithoutReference(Object proxy, String property, MetadataCache cache) 
        {
		    Object entity;
		    bool sureFromUs = false;
		    if ( proxy is INHibernateProxy ) 
            {
			    ILazyInitializer li = ( ( INHibernateProxy ) proxy ).HibernateLazyInitializer;
			    if ( li.IsUninitialized ) 
                {
				    return LoadState.NOT_LOADED;
			    }
			    else 
                {
				    entity = li.GetImplementation();
			    }
			    sureFromUs = true;
		    }
		    else 
            {
			    entity = proxy;
		    }
            
		    //we are instrumenting but we can't assume we are the only ones
		    if ( FieldInterceptionHelper.IsInstrumented( entity ) ) 
            {
			    IFieldInterceptor interceptor = FieldInterceptionHelper.ExtractFieldInterceptor( entity );
			    bool isInitialized = interceptor == null || interceptor.IsInitializedField( property );
			    LoadState state;
			    if (isInitialized && interceptor != null) 
                {
				    //property is loaded according to bytecode enhancement, but is it loaded as far as association?
				    //it's ours, we can read
				    state = isLoaded( get( entity, property, cache ) );
				    //it's ours so we know it's loaded
				    if (state == LoadState.UNKNOWN) state = LoadState.LOADED;
			    }
			    else if ( interceptor != null && (! isInitialized)) 
                {
				    state = LoadState.NOT_LOADED;
			    }
			    else if ( sureFromUs ) //interceptor == null
                { 
				    //property is loaded according to bytecode enhancement, but is it loaded as far as association?
				    //it's ours, we can read
				    state = isLoaded( get( entity, property, cache ) );
				    //it's ours so we know it's loaded
				    if (state == LoadState.UNKNOWN) state = LoadState.LOADED;
			    }
			    else 
                {
				    state = LoadState.UNKNOWN;
			    }

			    return state;
		    }
		    else 
            {
			    //can't do sureFromUs ? LoadState.LOADED : LoadState.UNKNOWN;
			    //is that an association?
			    return LoadState.UNKNOWN;
		    }
	    }

	    public static LoadState IsLoadedWithReference(Object proxy, String property, MetadataCache cache) 
        {
		    //for sure we don't instrument and for sure it's not a lazy proxy
		    Object obj = get(proxy, property, cache);
		    return isLoaded( obj );
	    }

	    private static Object get(Object proxy, String property, MetadataCache cache) 
        {
		    Type type = proxy.GetType();

		    try {
			    MemberInfo memberInfo = cache.getMember( type, property );
			    if (memberInfo is FieldInfo) 
                {
				    return ( (FieldInfo) memberInfo ).GetValue( proxy );
			    }
			    else if (memberInfo is MethodInfo) 
                {
				    return ( (MethodInfo) memberInfo ).Invoke( proxy, null);
			    }
			    else 
                {
				    throw new AssertionFailure( "member object neither Field nor MethodInfo: " + memberInfo);
			    }
		    }
		    catch (FieldAccessException e ) 
            {
			    throw new PersistenceException( "Unable to access field or method: "
							    + type + "#"
							    + property, e);
		    }
            catch(MethodAccessException e)
            {
                throw new PersistenceException( "Unable to access method: "
							    + type + "#"
							    + property, e);
            }
		    catch ( TargetInvocationException e ) 
            {
			    throw new PersistenceException( "Unable to access field or method: "
							    + type + "#"
							    + property, e);
		    }
	    }

	    public static LoadState isLoaded(Object o) 
        {
		    if ( o is INHibernateProxy ) 
            {
			    bool isInitialized = !( ( INHibernateProxy ) o ).HibernateLazyInitializer.IsUninitialized;
			    return isInitialized ? LoadState.LOADED : LoadState.NOT_LOADED;
		    }
		    else if ( o is IPersistentCollection ) 
            {
			    bool isInitialized = ( ( IPersistentCollection ) o ).WasInitialized;
			    return isInitialized ? LoadState.LOADED : LoadState.NOT_LOADED;
		    }
		    else 
            {
			    return LoadState.UNKNOWN;
		    }
	    }

    }

    /**
	* Cache hierarchy and MemberInfo resolution 
	*/
	public class MetadataCache : ISerializable
    {
        [NonSerialized]
		private Dictionary<string, WeakReference> m_classCache = new Dictionary<string, WeakReference>();

        public void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            m_classCache = new Dictionary<string, WeakReference>();
        }

		public MemberInfo getMember(Type clazz, String property) 
        {
            ClassCache cache = null;
            WeakReference wr;
            m_classCache.TryGetValue(clazz.FullName, out wr);
            if (wr != null)
            {
                cache = wr.Target as ClassCache;
            }
            
			if (cache == null) 
            {
				cache = new ClassCache(clazz);
				m_classCache.Add(clazz.FullName,  new WeakReference(cache));
			}
            
            Dictionary<String, MemberInfo> members = cache.getMembers();
            MemberInfo member;
            members.TryGetValue(property, out member);
			if ( member == null ) 
            {
				member = findMember( clazz, property );
				cache.getMembers().Add( property, member );
			}
			return member;
		}

		private MemberInfo findMember(Type clazz, String property) 
        {
			List<Type> classes = getClassHierarchy( clazz );

			foreach (Type current in classes) 
            {
				FieldInfo field;
				
				field = current.GetField( property );
                if (field != null)
                {
                    return field;
                }
				
                MethodInfo method = getMethod(clazz, property);
                if (method != null)
                {
					return method;
				}
			}
			//we could not find any match
			throw new PersistenceException( "Unable to find field or MethodInfo: "
							+ clazz + "#"
							+ property);
		}

        /**
	     * Returns the MethodInfo with the specified name or <code>null</code> if it does not exist.
	     *
	     * @param clazz The class to check.
	     * @param MethodInfoName The MethodInfo name.
	     *
	     * @return Returns the MethodInfo with the specified name or <code>null</code> if it does not exist.
	     */
        private static MethodInfo getMethod(Type clazz, String methodName)
        {
            char[] str = methodName.ToCharArray();
            str[0] = Char.ToUpper(str[0]);
            methodName = new String(str);
            MethodInfo method = clazz.GetMethod("get" + methodName);
            if (method == null)
            {
                return clazz.GetMethod("is" + methodName);
            }
            return method;
        }

		private List<Type> getClassHierarchy(Type clazz) 
        {
			ClassCache cache = null;
            WeakReference wr;
            m_classCache.TryGetValue(clazz.FullName, out wr);
            if (wr != null)
            {
                cache = wr.Target as ClassCache;
            }
			if (cache == null) 
            {
				cache = new ClassCache(clazz);
				m_classCache.Add( clazz.FullName, new WeakReference( cache ));
			}
			return cache.getClassHierarchy();
		}

	}

    public class ClassCache
    {
        List<Type> m_classHierarchy;
        Dictionary<String, MemberInfo> m_members = new Dictionary<String, MemberInfo>();

        public ClassCache(Type clazz)
        {
            m_classHierarchy = findClassHierarchy(clazz);
        }

        private static List<Type> findClassHierarchy(Type clazz)
        {
            List<Type> classes = new List<Type>();
            Type current = clazz;
            do
            {
                classes.Add(current);
                current = current.BaseType;
            }
            while (current != null);
            return classes;
        }

        public Dictionary<String, MemberInfo> getMembers()
        {
            return m_members;
        }

        public List<Type> getClassHierarchy()
        {
            return m_classHierarchy;
        }

    }

}
